# Automatically generated, do not edit!

"""
What:    Python simulation
Author:  ingolf
Contact: arne.t.elve(at)ntnu.no
Date:    2019-06-13 16:43:32
Model:   cc_HEX_single_nodes
Case:    latest
"""

# Import packages:
import numpy as np                                   # Numerical python library
from scipy.integrate import ode                           # Integrator in scipy
import matplotlib.pyplot as plt                             # Data illustration
from funcUtils import IndexSet, khatriRao, blockReduce, blockProduct   # Custom

from constants_use import *                                  # Import all constants
from networks import *                           # Import all network variables
# from selections_ import *                           # Import equation selections
# from initial_states_ import *
# =================================== BODY ================================== #

# INDEX SETS:
N = IndexSet('node', mapping = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], blocking = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
A = IndexSet('arc', mapping = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], blocking = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
S = IndexSet('species', mapping = [0, 1, 2], blocking = [1, 1, 1])
N_x_S = IndexSet('node & species', mapping = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], blocking = [1, 1, 1, 1, 1, 2, 2, 2, 2, 2])
A_x_S = IndexSet('arc & species', mapping = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10], blocking = [1, 1, 1, 1, 2, 2, 2, 2, 0, 0, 0])
R = IndexSet('species_conversion', mapping = [0], blocking = [1])
N_x_R = IndexSet('node & species_conversion', mapping = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10], blocking = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0])

# Constant equations:
T_ref = np.multiply(en, T_298)
h_ref = np.transpose(np.dot(np.transpose(h_0), P_node_species))
cp = np.transpose(np.dot(np.transpose(cpn), P_node_species))
Pnsas = np.dot(np.transpose(P_node_species), P_arc_species)
Fn = khatriRao(F_mass_volumetric, [N, A], Pnsas, [N_x_S, A_x_S])


# INTEGRATING FUNCTION:
def derivative(t, n, H):
  """
  t: time
  state: n, H
  Integrating function:
  dxdt = derivative(t, state)

  Note: sequence of variables depends on integrator. Using scipy's ode requires
  time before state, while odeint needs state before time.  Also the integrator
  want  flat vectors,  no column vectors.  I, therefore, transpose  the vectors
  manually. Normal transpose is not sufficient.
  """
  # EQUATIONS:
  # print(n)
  H_ref = blockReduce(h_ref, S, N_x_S, n)
  Cp = blockReduce(cp, S, N_x_S, n)
  T = np.add((np.multiply((np.reciprocal(Cp)), (np.subtract(H, H_ref)))), T_ref)
  qhat = np.multiply((np.multiply((-(U)), A_cross)), (np.dot(np.transpose(F_energy_heat), T)))
  ntot = blockReduce(ens, S, N_x_S, n)
  p = np.multiply((np.multiply((np.multiply((np.reciprocal(Vg)), ntot)), T)), Rg)
  dir = np.sign(np.dot(np.transpose(F_mass_volumetric), p))
  s_flow = np.multiply(half, (np.add((np.abs(F_mass_volumetric )), (np.multiply(F_mass_volumetric, np.transpose(dir))))))
  Vhat = np.multiply((np.multiply((-(kappa)), A_cross)), (np.dot(np.transpose(F_mass_volumetric), p)))
  Hhat = np.multiply(Vhat, (np.dot(np.transpose(s_flow), (np.multiply((np.reciprocal(Vg)), H)))))
  Hdot = np.add((np.dot(F_mass_volumetric, Hhat)), (np.dot(F_energy_heat, qhat)))
  c = khatriRao(np.reciprocal(Vg), [N], n, [N_x_S])
  chat = np.dot(np.transpose((khatriRao(s_flow, [N, A], Pnsas, [N_x_S, A_x_S]))), c)
  nhat = khatriRao(Vhat, [A], chat, [A_x_S])
  ndot = np.dot(Fn, nhat)

  a = np.transpose(np.array([[1.]*15]))
  a[0,0] = 0.
  a[4,0] = 0.
  a[5,0] = 0.
  a[6,0] = 0.
  a[13,0] = 0.
  a[14,0] = 0.
  # print(a, ndot)

  b = np.transpose(np.array([[1.]*10]))
  b[0,0] = 0.
  b[4,0] = 0.
  b[5,0] = 0.
  b[9,0] = 0.

  # FIX THIS:
  ndot_ret = np.multiply(ndot, a)
  Hdot_ret = np.multiply(Hdot, b)
  return ndot_ret, Hdot_ret

def integrand(t, state):
  # HACK: Integrator give flat state vectors while our model is column vector
  n = state[0:15, np.newaxis]
  H = state[15:25, np.newaxis]
  ndot, Hdot = derivative(t, n,  H)
  return np.concatenate([np.transpose(ndot)[0], np.transpose(Hdot)[0]])

# INTEGRATOR
data = {}
data['n'] = []
data['H'] = []
data['t'] = []
dt = 0.1
t_start = 0
t_end = 10.005                                               # Handcoded!!!
integrator = ode(integrand).set_integrator('dop853')
state = np.concatenate([np.transpose(n)[0], np.transpose(H)[0]])
integrator.set_initial_value(state, t_start)
data['t'].append(integrator.t)
data['n'].append(state[0:15, np.newaxis])
data['H'].append(state[15:25, np.newaxis])
while integrator.successful() and integrator.t < t_end:
  state = integrator.integrate(integrator.t + dt)
  print(integrator.t)
  data['t'].append(integrator.t)
  data['n'].append(state[0:15, np.newaxis])
  data['H'].append(state[15:25, np.newaxis])


data['n'] = np.transpose(data['n'])[0]
for dat in data['n']:
  plt.plot(data['t'], dat)
plt.xlim(0,10)
plt.ylim(0,16000)
plt.xlabel('Time [s]')
plt.ylabel('Number of moles [n]')
plt.legend(['A_h_in','A_h_1','A_h_2''A_h_3','A_h_out','B_c_in','C_c_in','B_c_1','C_c_1','B_c_2','C_c_2','B_c_3','C_c_3','B_c_out','C_c_out'])
plt.show()

data['H'] = np.transpose(data['H'])[0]
for i, (dat) in enumerate(data['H']):
  plt.plot(data['t'], dat)
plt.xlim(0,10)
plt.ylim(0,160000000)
plt.xlabel('Time [s]')
plt.ylabel('Enthalpy [H]')
plt.legend(['h_in','h_1','h_2','h_3','h_out','c_in','c_1','c_2','c_3','c_out'])
plt.show()
